# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path

from colcon_core.logging import colcon_logger
from colcon_core.package_discovery import PackageDiscoveryExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_metadata.metadata import get_metadata_files
from colcon_metadata.metadata import get_metadata_path
from colcon_metadata.metadata import metadata_by_name
from colcon_metadata.metadata import metadata_by_path
import yaml

logger = colcon_logger.getChild(__name__)


class ColconMetadataDiscovery(PackageDiscoveryExtensionPoint):
    """Provide meta information from `.meta` files."""

    # the priority needs to be higher than other extensions
    # in order for metadata to be added before those are invoked
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageDiscoveryExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser, with_default):  # noqa: D102
        parser.add_argument(
            '--ignore-user-meta',
            action='store_true',
            help='Ignore `*.meta` files in the user config directory `%s`' %
                 get_metadata_path())
        parser.add_argument(
            '--metas',
            nargs='*',
            metavar='PATH',
            default='./colcon.meta',
            help='The directories containing a `colcon.meta` file or paths to '
                 'arbitrary files containing the same meta information '
                 '(default: ./colcon.meta)')

    def has_parameters(self, *, args):  # noqa: D102
        if not args.ignore_user_meta:
            for path in get_metadata_files():
                self._add(Path(path))

        for meta in args.metas or []:
            path = Path(meta)
            if not path.exists():
                continue
            if path.is_dir():
                path_file = path / 'colcon.meta'
                if not path_file.exists():
                    continue
                path = path_file
            self._add(path)

        # doesn't implement the `discover()` method
        # since `has_parameters()` always returns False
        return False

    def _add(self, meta_path):
        content = meta_path.read_text()
        try:
            data = yaml.load(content)
        except Exception as e:
            logger.warn(
                "Skipping metadata file '%s' since it failed to parse: %s" %
                (meta_path.absolute(), e))
            return

        if data is None:
            logger.info(
                "Empty metadata file '%s'" % meta_path.absolute())
            return
        if not isinstance(data, dict):
            logger.warn(
                "Skipping metadata file '%s' since it doesn't contain a dict" %
                meta_path.absolute())
            return

        logger.info(
            "Using configuration from '%s'" % meta_path.absolute())
        if 'names' in data and isinstance(data['names'], dict):
            for name, v in data['names'].items():
                metadata_by_name[name].update(v)
        if 'paths' in data and isinstance(data['paths'], dict):
            for path, value in data['paths'].items():
                path = meta_path.parent / path
                metadata_by_path[
                    os.path.realpath(str(path))].update(value)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import defaultdict
import os

from colcon_core.location import get_config_path

metadata_by_name = defaultdict(dict)
metadata_by_path = defaultdict(dict)


def get_metadata_path():
    """
    Get the path where metadata is stored.

    :rtype: Path
    """
    return get_config_path() / 'metadata'


def get_metadata_files(path=None):
    """
    Get the paths of all metadata files.

    The metadata path is recursively being crawled for files ending in `.meta`.
    Directories starting with a dot (`.`) are being ignored.

    :rtype: list
    """
    metadata_path = path or get_metadata_path()
    if not metadata_path.is_dir():
        return []

    files = []
    for dirpath, dirnames, filenames in os.walk(
        str(metadata_path), followlinks=True
    ):
        # skip subdirectories starting with a dot
        dirnames[:] = filter(lambda d: not d.startswith('.'), dirnames)
        dirnames.sort()

        for filename in sorted(filenames):
            if not filename.endswith('.meta'):
                continue
            path = os.path.join(dirpath, filename)
            files.append(path)
    return files



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_augmentation import update_descriptor
from colcon_core.package_identification import logger
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_metadata.metadata import metadata_by_name
from colcon_metadata.metadata import metadata_by_path


class ColconMetadataPackageIdentification(
    PackageIdentificationExtensionPoint, PackageAugmentationExtensionPoint,
):
    """Provide meta information from `.colcon` files."""

    # the priority needs to be higher than other extensions in order to
    # populate the descriptor with meta data before those are invoked
    PRIORITY = 180

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def identify(self, desc):  # noqa: D102
        if str(desc.path) in metadata_by_path:
            metadata2 = metadata_by_path[str(desc.path)]
            if metadata2.get('type'):
                if desc.type is None:
                    desc.type = metadata2.get('type')
                    logger.info(
                        "Package '{desc.path}' has configured type "
                        "'{desc.type}'".format_map(locals()))
                elif desc.type != metadata2.get('type'):
                    logger.warn(
                        "type '%s' of package '%s' does not match configured "
                        "type '%s'", desc.type, desc.path,
                        metadata2.get('type'))
            if metadata2.get('name'):
                if desc.name is None:
                    desc.name = metadata2.get('name')
                    logger.info(
                        "Package '{desc.path}' has configured name "
                        "'{desc.name}'" .format_map(locals()))
                elif desc.name != metadata2.get('name'):
                    logger.warn(
                        "name '%s' of package '%s' does not match configured "
                        "name '%s'", desc.name, desc.path,
                        metadata2.get('name'))

        if desc.name in metadata_by_name:
            data = metadata_by_name[desc.name]
            if 'type' in data:
                # if the package name was defined based on the path
                # then ignore the type defined for the package name
                if desc.type is not None and desc.type != data['type']:
                    logger.warn(
                        "type '{desc.type}' of package '{desc.name}' "
                        "does not match configured type '{data[type]}'"
                        .format_map(locals()))
                else:
                    desc.type = data['type']

    def augment_package(
        self, desc, *, additional_argument_names
    ):  # noqa: D102
        # deliberatively ignore the package type
        # since this extension can contribute meta information to any package
        if desc.name in metadata_by_name:
            update_descriptor(
                desc, metadata_by_name[desc.name],
                additional_argument_names=additional_argument_names)
        if str(desc.path) in metadata_by_path:
            update_descriptor(
                desc, metadata_by_path[str(desc.path)],
                additional_argument_names=additional_argument_names)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import json

from colcon_core.logging import colcon_logger
from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_augmentation import update_descriptor
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version

logger = colcon_logger.getChild(__name__)


class ColconPkgPackageIdentification(
    PackageIdentificationExtensionPoint, PackageAugmentationExtensionPoint,
):
    """Provide meta information from `colcon.pkg` files."""

    # the priority needs to be higher than other extensions in order to
    # populate the descriptor with meta data before those are invoked
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def identify(self, desc):  # noqa: D102
        data = self._get_colcon_pkg_data(desc.path)
        if not data:
            return

        # check for package type
        if 'type' in data:
            if desc.type is not None and desc.type != data['type']:
                logger.warn(
                    "Type '%s' of package '%s' does not match configured "
                    "type '%s'", desc.type, desc.path, data['type'])
            else:
                logger.info(
                    "Setting type '%s' of package '%s' from colcon.pkg",
                    data['type'], desc.path)
                desc.type = data['type']

        # check for package name
        if 'name' in data:
            if desc.name is not None and desc.name != data['name']:
                logger.warn(
                    "Name '%s' of package '%s' does not match configured "
                    "name '%s'", desc.name, desc.path, data['name'])
            else:
                logger.info(
                    "Setting name '%s' of package '%s' from colcon.pkg",
                    data['name'], desc.path)
                desc.name = data['name']

    def augment_package(
        self, desc, *, additional_argument_names
    ):  # noqa: D102
        data = self._get_colcon_pkg_data(desc.path)
        if not data:
            return

        logger.info(
            "Updating descriptor of package '%s' from colcon.pkg: %s",
            desc.path, data)
        update_descriptor(
            desc, data, additional_argument_names=additional_argument_names)

    def _get_colcon_pkg_data(self, basepath):
        # check for a colcon.pkg file in the root of the package directory
        colcon_pkg_path = basepath / 'colcon.pkg'
        if not colcon_pkg_path.is_file():
            return None
        content = colcon_pkg_path.read_text()
        return json.loads(content)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.plugin_system import satisfies_version
from colcon_metadata.metadata.repository import get_repositories
from colcon_metadata.metadata.repository import get_repository_metadata_files
from colcon_metadata.subverb import MetadataSubverbExtensionPoint


class ListMetadataSubverb(MetadataSubverbExtensionPoint):
    """List all repositories and their metadata."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            MetadataSubverbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        argument = parser.add_argument(
            'name',
            nargs='?',
            help='Only list the information for a specific repository')
        try:
            from argcomplete.completers import ChoicesCompleter
        except ImportError:
            pass
        else:
            repos = get_repositories()
            argument.completer = \
                ChoicesCompleter(repos.keys())

    def main(self, *, context):  # noqa: D102
        repos = get_repositories()
        if context.args.name and context.args.name not in repos.keys():
            return "Passed repository name '{context.args.name}' is unknown" \
                .format_map(locals())

        for name in sorted(repos.keys()):
            if context.args.name and context.args.name != name:
                continue
            url = repos[name]
            print('{name}: {url}'.format_map(locals()))
            metadata_files = get_repository_metadata_files(
                repository_name=name)
            for path in sorted(metadata_files):
                print('- {path}'.format_map(locals()))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import socket
import time
from urllib.error import HTTPError
from urllib.error import URLError
from urllib.request import urlopen

from colcon_core.location import get_config_path
from colcon_core.logging import colcon_logger
from colcon_metadata.metadata import get_metadata_files
from colcon_metadata.metadata import get_metadata_path
import yaml

logger = colcon_logger.getChild(__name__)

"""The path of the yaml file describing the metadata repositories."""
metadata_repositories_file = get_config_path() / 'metadata_repositories.yaml'


def get_repositories():
    """
    Get the registered repositories.

    :rtype: dict
    """
    global metadata_repositories_file
    if not metadata_repositories_file.exists():
        return {}
    if metadata_repositories_file.is_dir():
        raise IsADirectoryError()
    content = metadata_repositories_file.read_text()
    data = yaml.load(content)
    assert isinstance(data, dict), 'The content of the configuration file ' \
        "'%s' should be a dictionary" % metadata_repositories_file
    return data


def set_repositories(repositories):
    """
    Persist the passed repositories in the configuration file.

    :param dict repositories: The repositories
    """
    assert isinstance(repositories, dict), \
        'The passed repositories should be a dictionary'
    global metadata_repositories_file
    data = yaml.dump(repositories, default_flow_style=False)
    os.makedirs(str(metadata_repositories_file.parent), exist_ok=True)
    with metadata_repositories_file.open('w') as h:
        h.write(data)


def get_repository_metadata_files(*, repository_name):
    """
    Get the configuration files for a specific repository.

    :param str repository_name: The repository name
    :rtype: list
    """
    return get_metadata_files(get_metadata_path() / repository_name)


def load_url(url, retry=2, retry_period=1, timeout=10):
    """
    Load a URL.

    :param int retry: The number of retries in case the request fails
    :param int retry_period: The period to wait before the first retry. Every
      subsequent retry will double the period.
    :param int timeout: The timeout for each request

    :rtype: str
    """
    try:
        h = urlopen(url, timeout=timeout)
    except HTTPError as e:
        if e.code == 503 and retry:
            time.sleep(retry_period)
            return load_url(
                url, retry=retry - 1, retry_period=retry_period * 2,
                timeout=timeout)
        e.msg += ' (%s)' % url
        raise
    except URLError as e:
        if isinstance(e.reason, socket.timeout) and retry:
            time.sleep(retry_period)
            return load_url(
                url, retry=retry - 1, retry_period=retry_period * 2,
                timeout=timeout)
        raise URLError(str(e) + ' (%s)' % url)
    except socket.timeout as e:
        if retry:
            time.sleep(retry_period)
            return load_url(
                url, retry=retry - 1, retry_period=retry_period * 2,
                timeout=timeout)
        raise socket.timeout(str(e) + ' (%s)' % url)
    content = h.read()
    return content.decode('utf-8')



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import sys

from colcon_core.plugin_system import satisfies_version
from colcon_metadata.metadata import get_metadata_path
from colcon_metadata.metadata.repository import get_repositories
from colcon_metadata.metadata.repository import get_repository_metadata_files
from colcon_metadata.metadata.repository import load_url
from colcon_metadata.subverb import MetadataSubverbExtensionPoint
import yaml


class UpdateMetadataSubverb(MetadataSubverbExtensionPoint):
    """Update the metadata from the repository indexes."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            MetadataSubverbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.description += '\n\n' \
            'For each repository all metadata files are being fetched. ' \
            'The status if each metadata file is indicated by the following ' \
            'symbols:\n' \
            '  + added new metadata file\n' \
            '  * updated the existing metadata file\n' \
            '  . existing metadata file was already the same\n' \
            '  - renamed obsolete metadata file'
        argument = parser.add_argument(
            'name',
            nargs='?',
            help='Only update the metadata from a specific repository')
        try:
            from argcomplete.completers import ChoicesCompleter
        except ImportError:
            pass
        else:
            repos = get_repositories()
            argument.completer = \
                ChoicesCompleter(repos.keys())

    def main(self, *, context):  # noqa: D102
        repos = get_repositories()
        if context.args.name and context.args.name not in repos.keys():
            return "Passed repository name '{context.args.name}' is unknown" \
                .format_map(locals())

        # IDEA fetch all resources in parallel
        rc = 0
        for name in sorted(repos.keys()):
            if context.args.name and context.args.name != name:
                continue

            # fetch the repository index
            index_url = repos[name]
            print('fetching {name}: {index_url} ...'.format_map(locals()))
            try:
                content = load_url(index_url)
            except Exception as e:
                print(' ', str(e), file=sys.stderr)
                rc = 1
                continue

            # parse the repository index
            try:
                data = yaml.load(content)
            except Exception as e:
                print(' ', str(e), file=sys.stderr)
                rc = 1
                continue
            if not isinstance(data, dict) or 'metadata' not in data.keys():
                print('  The repository index should be a dictionary with a '
                      "'metadata' key, but it is: {data}".format_map(locals()))
                rc = 1
                continue

            # get existing metadata files to remove obsolete ones later
            metadata_files_before = get_repository_metadata_files(
                repository_name=name)

            # fetch all metadata files referenced in the index
            metadata_urls = data['metadata']
            metadata_basenames = set()
            for metadata_url in metadata_urls:
                # if metadata URL is relative prefix the dirname of the index
                if (
                    '://' not in metadata_url and
                    not os.path.isabs(metadata_url)
                ):
                    metadata_url = \
                        os.path.dirname(index_url) + '/' + metadata_url

                # fetch the metadata file
                print('  fetching {metadata_url} ...'.format_map(locals()))
                try:
                    content = load_url(metadata_url)
                except Exception as e:
                    print('  -', str(e), file=sys.stderr)
                    rc = 1
                    continue

                # save the metadata file
                metadata_basename = os.path.basename(metadata_url)
                if metadata_basename in metadata_basenames:
                    print('  Multiple metadata files with the same basename '
                          "'{metadata_basename}'".format_map(locals()),
                          file=sys.stderr)
                else:
                    metadata_basenames.add(metadata_basename)

                destination_basepath = get_metadata_path() / name
                os.makedirs(str(destination_basepath), exist_ok=True)
                destination_path = destination_basepath / metadata_basename
                if not destination_path.exists():
                    mod = '+'
                else:
                    if content == destination_path.read_text():
                        mod = '.'
                    else:
                        # IDEA show the diff if the file already exists
                        mod = '*'
                print(' ', mod, str(destination_path))
                with destination_path.open('w') as h:
                    h.write(content)

            # remove obsolete metadata files
            for metadata_file in metadata_files_before:
                if os.path.basename(metadata_file) not in metadata_basenames:
                    os.rename(metadata_file, metadata_file + '.obsolete')
                    print('  - {metadata_file} -> *.obsolete'
                          .format_map(locals()))

        return rc



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name

logger = colcon_logger.getChild(__name__)


class MetadataSubverbExtensionPoint:
    """
    The interface for metadata subverb extensions.

    A metadata subverb extension provides a subverb to the `metadata` verb of
    the command line tool.

    For each instance the attribute `SUBVERB_NAME` is being set to the basename
    of the entry point registering the extension.
    """

    """The version of the metadata subverb extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the metadata subverb.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def main(self, *, context):
        """
        Execute the metadata subverb extension logic.

        This method must be overridden in a subclass.

        :param context: The context providing the parsed command line arguments
        :returns: The return code
        """
        raise NotImplementedError()


def get_subverb_extensions():
    """
    Get the available subverb extensions.

    The extensions are ordered by their entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.SUBVERB_NAME = name
    return order_extensions_by_name(extensions)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from argparse import ArgumentTypeError

from colcon_core.plugin_system import satisfies_version
from colcon_metadata.metadata.repository import get_repositories
from colcon_metadata.metadata.repository import set_repositories
from colcon_metadata.subverb import MetadataSubverbExtensionPoint


class AddMetadataSubverb(MetadataSubverbExtensionPoint):
    """Add the URL of a repository index."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            MetadataSubverbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            'name',
            type=_non_empty_string_without_pathsep,
            help='The unique name identifying the repository')
        parser.add_argument(
            'url',
            type=_url_string,
            help='The url of a metadata repository index')

    def main(self, *, context):  # noqa: D102
        repos = get_repositories()
        if context.args.name in repos.keys():
            return "A repository with the name '{context.args.name}' " \
                'already exists'.format_map(locals())
        repos[context.args.name] = context.args.url
        set_repositories(repos)


def _non_empty_string_without_pathsep(value):
    if not value:
        raise ArgumentTypeError('must be a non-empty string')
    for pathsep in ('/', '\\'):
        if pathsep in value:
            raise ArgumentTypeError(
                "must not contain '{pathsep}'".format_map(locals()))
    return value


def _url_string(value):
    if '://' not in value:
        raise ArgumentTypeError("must contain '://'")
    return value
